# cython: language_level=3
# cython: cdivision=True
from cpython.bytes cimport PyBytes_AS_STRING, PyBytes_FromStringAndSize
from libc.stdint cimport uint8_t

from pysmaz.backends.cython.smaz cimport smaz_compress, smaz_decompress


cpdef inline bytes compress(const uint8_t[::1] data, Py_ssize_t output_size = 0):
    cdef Py_ssize_t input_len = data.shape[0]
    if output_size==0:
        output_size = input_len # 如果没有指定输出buffer的大小，默认就是2倍输入大小
    cdef bytes output = PyBytes_FromStringAndSize(NULL, output_size)
    if <void*>output == NULL:
        raise MemoryError
    cdef int buffer_updated
    cdef char *output_ptr = PyBytes_AS_STRING(output)
    with nogil:
        buffer_updated = smaz_compress(<char*>&data[0], <int>input_len, output_ptr, <int>output_size)
    if buffer_updated == <int>input_len+1:
        raise ValueError("output buffer is too small")
    
    return output[:buffer_updated]

cpdef inline bytes decompress(const uint8_t[::1] data, Py_ssize_t output_size = 0):
    cdef Py_ssize_t input_len = data.shape[0]
    if output_size==0:
        output_size = input_len * 2 # 如果没有指定输出buffer的大小，默认就是2倍输入大小
    
    cdef bytes output = PyBytes_FromStringAndSize(NULL, output_size)
    if <void*>output == NULL:
        raise MemoryError
    cdef int buffer_updated
    cdef char *output_ptr = PyBytes_AS_STRING(output)
    with nogil:
        buffer_updated = smaz_decompress(<char*>&data[0], <int>input_len, output_ptr, <int>output_size)
    if buffer_updated == <int>output_size+1:
        raise ValueError("output buffer is too small")

    return output[:buffer_updated]


cpdef inline int compress_into(const uint8_t[::1] data, uint8_t[::1] output) except? 0:
    cdef Py_ssize_t output_size = output.shape[0]
    cdef int ret
    with nogil:
        ret =  smaz_compress(<char *> &data[0], <int> data.shape[0], <char*>&output[0], <int>output_size)
    if ret == <int>output_size + 1:
        raise ValueError("output buffer is too small")
    return ret

cpdef inline int decompress_into(const uint8_t[::1] data, uint8_t[::1] output) except? 0:
    cdef Py_ssize_t output_size = output.shape[0]
    cdef int ret
    with nogil:
        ret =  smaz_decompress(<char *> &data[0], <int> data.shape[0], <char*>&output[0], <int>output_size)
    if ret == <int>output_size + 1:
        raise ValueError("output buffer is too small")
    return ret


"""
Copyright (c) 2008-2021 synodriver <synodriver@gmail.com>
"""
import sys
from random import choice, randint
from unittest import TestCase

sys.path.append(".")
import os

os.environ["SMAZ_USE_CFFI"] = "1"

from pysmaz import compress, decompress


class TestAll(TestCase):
    def setUp(self) -> None:
        pass

    def test_raise(self):
        with self.assertRaises(ValueError):
            compress(b"1234")

    def test_encode(self):
        for i in range(1000):
            data = bytes([randint(0, 255) for _ in range(randint(100, 1000))])
            out = compress(data, 3000)
            self.assertEqual(decompress(out, 2000), data)
        pass

    def tearDown(self) -> None:
        pass


if __name__ == "__main__":
    import unittest

    unittest.main()



# cython: language_level=3
# cython: cdivision=True
cdef extern from "smaz.h" nogil:
    int smaz_compress(char *in_, int inlen, char *out, int outlen)
    int smaz_decompress(char *in_, int inlen, char *out, int outlen)


import glob
import sys

from cffi import FFI

ffibuilder = FFI()
ffibuilder.cdef(
    """
int smaz_compress(char *in, int inlen, char *out, int outlen);
int smaz_decompress(char *in, int inlen, char *out, int outlen);
    """
)

source = """
#include "smaz.h"
"""

ffibuilder.set_source(
    "pysmaz.backends.cffi._smaz",
    source,
    sources=["./smaz/smaz.c"],
    include_dirs=["./smaz"],
)

if __name__ == "__main__":
    ffibuilder.compile()



"""
Copyright (c) 2008-2021 synodriver <synodriver@gmail.com>
"""
from pysmaz.backends.cffi._smaz import ffi, lib


def compress(data: bytes, output_size=0) -> bytes:
    input_len = len(data)
    if output_size == 0:
        output_size = input_len  # 如果没有指定输出buffer的大小，默认就是就是输入大小
    output = ffi.new(f"char[{output_size}]")

    buffer_updated: int = lib.smaz_compress(
        ffi.from_buffer(data), input_len, output, output_size
    )
    if buffer_updated == output_size + 1:
        raise ValueError("output buffer is too small")
    return ffi.unpack(output, buffer_updated)


def decompress(data: bytes, output_size=0) -> bytes:
    input_len = len(data)
    if output_size == 0:
        output_size = input_len * 2  # 如果没有指定输出buffer的大小，默认就是2倍输入大小
    output = ffi.new(f"char[{output_size}]")

    buffer_updated: int = lib.smaz_decompress(
        ffi.from_buffer(data), input_len, output, output_size
    )
    if buffer_updated == output_size + 1:
        raise ValueError("output buffer is too small")
    return ffi.unpack(output, buffer_updated)


def compress_into(data: bytes, output: bytearray) -> int:
    output_size = len(output)
    ret: int = lib.smaz_compress(
        ffi.from_buffer(data), len(data), ffi.from_buffer(output), output_size
    )
    if ret == output_size + 1:
        raise ValueError("output buffer is too small")
    return ret


def decompress_into(data: bytes, output: bytearray) -> int:
    output_size = len(output)
    ret: int = lib.smaz_decompress(
        ffi.from_buffer(data), len(data), ffi.from_buffer(output), output_size
    )
    if ret == output_size + 1:
        raise ValueError("output buffer is too small")
    return ret
